(* 	      On Graph Rewriting, Reduction and Evaluation.
 *			      Ian Zerny
 *	                   Copyright 2009
 *                GPLv3, see LICENSE.txt for details
 *
 *		      www.zerny.dk/on-graph-rewriting.html
 *
 *			Version: November 2009
 *)

use "ski.sml";
use "ba.sml";
(* use "ho.sml"; included in ski.sml *)

(* ============================================================
 * Testing Functor.
 * ============================================================ *)

functor Test (M : SKI)
= struct
local open M in

(* composition = S(KS)K *)
fun b () = a (a (s()) (a (k()) (s()))) (k())

(* doubler = SII *)
fun d () = a (a (s()) (i())) (i())

(* omega = SII(SII) *)
fun omega () = a (d()) (d())

(* Create a balanced tree of I applications of depth n *)
fun itree 0 = (i())
  | itree n = let val t = itree (n-1)
              in a t t
              end

(* Reduce a term a number of steps. *)
fun steps 0 t = t
  | steps n t = steps (n-1) (reduce t)

(* Some tests *)
val z0 = compare (normalize (i()), (i()))
val z1 = compare (normalize (k()), (k()))
val z2 = compare (normalize (s()), (s()))
val z3 = compare (normalize (itree 1), (i()))
val z4 = compare (normalize (itree 5), (i()))

fun t5 () = a (d()) (k())
val r5 = normalize (t5 ())
val z5 = compare (r5, a (k()) (k()))

(* Test proper sharing (SII(KII)) -> I(II) in 4 steps *)
fun t6 () = sa (i()) (i()) (ka (i()) (i()))
val z6 = compare (steps 4 (t6()), ia (a (i()) (i())))

(* Test that exponential behavior is avoided *)
fun sapps 0 = i ()
  | sapps n = sa (i()) (i()) (sapps (n-1))
val z7 = compare (i(), normalize (sapps 100))


(* Some computational tests using bracket abstraction of lambda-terms. *)
local open BA in

exception TEST_ERROR of string
exception INVALID_BRACKET_ABSTRACTION

(*  ba2sk : BA.t -> M.t *)
fun ba2sk (VAR "I") = i ()
  | ba2sk (VAR "K") = k ()
  | ba2sk (VAR "S") = s ()
  | ba2sk (VAR  x ) = raise INVALID_BRACKET_ABSTRACTION
  | ba2sk (LAM (x, e)) = raise INVALID_BRACKET_ABSTRACTION
  | ba2sk (APP (e0, e1)) = a (ba2sk e0) (ba2sk e1)

(*  run : BA.t -> HO.t [hopefully a LIT n] *)
val run = HO.c2n o sk2ho o normalize o ba2sk o ba
(* NOTE: this definition of run embodies two reduction mechanisms:
 *       that of lambda-terms as defined by eval in ho.sml, and
 *       that of SKI-terms as defined by our reduction implementations.
 *)

(*  zro = λ f x -> x *)
val zro = LAM ("f", LAM ("x", VAR "x"))

(*  scc n = λ f x -> n f (f x) *)
val scc = LAM ("n", LAM ("f", LAM ("x",
               APP (APP (VAR "n", VAR "f"),
                    APP (VAR "f", VAR "x")))))

(*  plus n m = λ f x -> n f (m f x) *)
val plus = LAM ("n", LAM ("m", LAM ("f", LAM ("x",
                APP (APP (VAR "n", VAR "f"),
                     APP (APP (VAR "m", VAR "f"),
                          VAR "x"))))))

(*  mult n m = λ f x -> n (m f) x *)
val mult = LAM ("n", LAM ("m", LAM ("f",
                APP (VAR "n", APP (VAR "m", VAR "f")))))

(* The test terms *)
val one = APP (scc, zro)
val two = APP (scc, one)
val thr = APP (APP (plus, two), one)
val six = APP (APP (mult, two), thr)
val asw = APP (APP (plus, APP (APP (mult, six), six)), six)

(* The computational tests *)
fun zn () = (
    (case run one of HO.LIT  1 => true | _ => raise (TEST_ERROR "zn0"));
    (case run two of HO.LIT  2 => true | _ => raise (TEST_ERROR "zn1"));
    (case run thr of HO.LIT  3 => true | _ => raise (TEST_ERROR "zn2"));
    (case run six of HO.LIT  6 => true | _ => raise (TEST_ERROR "zn3"));
    (case run asw of HO.LIT 42 => true | _ => raise (TEST_ERROR "zn4")))

end (* close BA *)

(* Result of all tests *)
val z = List.all (fn x => x)
        [z0,z1,z2,z3,z4,z5,z6,z7,
         (zn () handle
          TEST_ERROR s => (print ("Test failed: "^s^"\n"); false))]

end
end;


(* ============================================================
 * Test instances.
 * ============================================================ *)

structure TestGRS0 = Test (GRS0);
structure TestGRS1 = Test (GRS1);
structure TestGRS2 = Test (GRS2);
structure TestGRS3 = Test (GRS3);
structure TestGRS4 = Test (GRS4);
structure TestRM1 = Test (RM1);
structure TestRM4 = Test (RM4);
structure TestRM5 = Test (RM5);
structure TestRM6 = Test (RM6);


(* Top-level test for easy viewing. *)
val z =
    (* GRS *)
    TestGRS0.z andalso
    TestGRS1.z andalso
    TestGRS2.z andalso
    TestGRS3.z andalso
    TestGRS4.z andalso
    (* RM *)
    TestRM1.z andalso
    TestRM4.z andalso
    TestRM5.z andalso
    TestRM6.z
